package antlr4_parser;

import java.util.ArrayList;
import java.util.List;

import org.antlr.v4.runtime.ANTLRInputStream;
import org.antlr.v4.runtime.CommonTokenStream;
import org.antlr.v4.runtime.misc.ParseCancellationException;
import org.antlr.v4.runtime.tree.ParseTree;

public class MethodParser {
	
  public static List<Method> parseString(String text) throws ParseCancellationException {
  	
  	ANTLRInputStream rawInputChars = new ANTLRInputStream(text);
    JavaMethodsLexer lexer = new JavaMethodsLexer(rawInputChars);
    wrapLexerErrorToException(lexer);
    CommonTokenStream tokens = new CommonTokenStream(lexer);
    JavaMethodsParser parser = new JavaMethodsParser(tokens);
    wrapParserErrorToException(parser);
    ParseTree parseTree = parser.methods();
    
		SignatureVisitor signatureVisitor = new SignatureVisitor();
    String signatures = signatureVisitor.visit(parseTree);
    NiceFormattingVisitor niceFormattingVisitor = new NiceFormattingVisitor();
    String htmlFormattings = niceFormattingVisitor.visit(parseTree);
  	
    List<Method> methods = new ArrayList<Method>();
  	text = normalizeText(text);  
  	String[] methodArray = text.split(";");
  	String[] signatureArray = signatures.split(";");
    String[] htmlFormattingArray = htmlFormattings.split(";");
  	for(int i = 0; i < methodArray.length; i++) {
  		if(methodArray[i].startsWith(" "))
  			methodArray[i] = methodArray[i].substring(1);
  		methods.add(new Method(methodArray[i], signatureArray[i], htmlFormattingArray[i], null, null));
  	}
  	
		return methods;
  }
  
  private static String normalizeText(String text) {
  	text = text.trim().replace("(", " ( ");
  	text = text.trim().replace(")", " ) ");
  	text = text.trim().replace(",", ", ");
  	text = text.trim().replace(";", " ; ");
  	text = text.trim().replace("[", " [ ");
  	text = text.trim().replace("]", " ] ");
  	text = text.trim().replaceAll("\\s+", " ");
  	text = text.trim().replace(" [ ", "[");
  	return text;
  }
  
  /*Wrap lexical errors into a ParseCancellationException instead of
  writing them on stderr*/
  private static void wrapLexerErrorToException(JavaMethodsLexer lexer) {
  	lexer.removeErrorListeners();
  	lexer.addErrorListener(ThrowingErrorListener.INSTANCE);
  }

  /*Wrap parser errors into a ParseCancellationException instead of
  writing them on stderr*/
  private static void wrapParserErrorToException(JavaMethodsParser parser) {
  	parser.removeErrorListeners();
  	parser.addErrorListener(ThrowingErrorListener.INSTANCE);
  }
  
  public static boolean checkSyntax(String text) {
  	try {
  		if(text.contains("[") || text.contains("]")) // disable array support
  			return false;
  		ANTLRInputStream rawInputChars = new ANTLRInputStream(text);
  		JavaMethodsLexer lexer = new JavaMethodsLexer(rawInputChars);
  		wrapLexerErrorToException(lexer);
  		CommonTokenStream tokens = new CommonTokenStream(lexer);
  		JavaMethodsParser parser = new JavaMethodsParser(tokens);
  		wrapParserErrorToException(parser);
  		ParseTree parseTree = parser.methods();
  		return parseTree.getChildCount() == 1;
  	}
  	catch(Exception e) {
  		return false;
  	}
  }
  
  public static MethodData getMethodData(String text) {
  	text = normalizeText(text);
  	MethodData methodData = new MethodData();

  	String[] preOpenRoundBracket = text.split("\\(")[0].split(" ");
  	String[] betweenRoundBrackets = text.split("\\(")[1].split("\\)")[0].split(" ");	
  	String[] postCloseRoundBracket = null; //exceptions are optional
  	if(text.split("\\)").length > 1)
  	  postCloseRoundBracket = text.split("\\)")[1].split(" ");
  		
  	methodData.setName(preOpenRoundBracket[preOpenRoundBracket.length - 1]);
  	String returnType = "";
  	for(int i = 0; i < preOpenRoundBracket.length - 1; i++)
  		returnType += preOpenRoundBracket[i];
  	methodData.setReturnType(returnType);

  	String parametersString = "";
  	for(int i = 0; i < betweenRoundBrackets.length; i++)
  		parametersString += betweenRoundBrackets[i] + " ";
  	parametersString = parametersString.trim().replace(", ", ",");
  	for(int i = 0; i < parametersString.split(",").length; i++)
  		if(!parametersString.split(",")[i].equals(""))
  		  methodData.addParameter(parametersString.split(",")[i]);

  	if(postCloseRoundBracket != null) {
  		String exceptionsString = "";
  		for(int i = 2; i < postCloseRoundBracket.length - 1; i++)
  			exceptionsString += postCloseRoundBracket[i];
  		for(int i = 0; i < exceptionsString.split(",").length; i++)
  			if(!exceptionsString.split(",")[i].equals(""))
  				methodData.addException(exceptionsString.split(",")[i]);
  	}

  	return methodData;
  }
}
